/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gmeBackend.h"

#ifdef BUILD_GME

#include <FX88591Codec.h>
#include "utils.h"

#ifdef HAVE_LIBZ
#  define EXT	"*.(ay|gbs|gym|hes|kss|nsfe|nsf|sap|spc|vgm|vgz)"
#else
#  define EXT	"*.(ay|gbs|gym|hes|kss|nsfe|nsf|sap|spc|vgm)"
#endif

// ASMA path to STIL.
#define ASMA_STIL	"/Docs/STIL.txt"

// ASMA path to BUGlist.
#define ASMA_BUGLIST	"/Docs/Bugs.txt"

#define CREDITS		"Game_Music_Emu library 0.5.5\nCopyright \302\251 Shay Green."
#define LINK		"http://code.google.com/p/game-music-emu/"

const char gmeBackend::name[]="Gme";

gmeConfig_t gmeBackend::_settings;

/*****************************************************************/

size_t gmeBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	if (_emu->track_ended())
		return 0;

	_emu->play((bufferSize>>1), (Music_Emu::sample_t*)buffer);
	return bufferSize;
}

/*****************************************************************/

bool gmeBackend::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString gmeBackend::ext() const { return FXString(EXT); }

gmeBackend::gmeBackend() :
	inputBackend(name),
	_emu(0),
	_stil(0) {

	loadSettings();
	openAsma(_settings.asmaPath);
}

gmeBackend::~gmeBackend() {

	close();

	delete _stil;
}

void gmeBackend::loadSettings() {

	_settings.samplerate=load("Samplerate", 44100);
	_settings.asmaPath=load("ASMA", FXString::null);
}

void gmeBackend::saveSettings() {

	if (compare(_settings.asmaPath, load("ASMA", FXString::null))) {
		FXTRACE((10, "Reloading ASMA from %s\n", _settings.asmaPath.text()));
		openAsma(_settings.asmaPath);
	}

	save("Samplerate", _settings.samplerate);
	save("ASMA", _settings.asmaPath);
}

bool gmeBackend::open(const FXString& fileName) {

	close();

	gme_type_t fileType;
	if (!check(gme_identify_file(fileName.text(), &fileType)))
		return false;

	FXTRACE((10, "System %s\n", fileType->system));

	_emu=fileType->new_emu();
	if (!check(_emu->set_sample_rate(_settings.samplerate)))
		return false;
	if (!check(_emu->load_file(fileName.text())))
		return false;
	if (!check(_emu->start_track(0)))
		return false;

	_emu->load_m3u(FXStringFormat("%s.m3u",FXPath::stripExtension(fileName).text()).text());

	getInfo();

	_hasStilInfo=_stil && !comparecase(FXPath::extension(fileName), "sap");
	if (_hasStilInfo) {
		FXTRACE((10, "Retrieving STIL info\n"));
		FXString comment=FXString(_stil->getAbsGlobalComment(fileName.text()));
		if (!comment.empty())
			comment.append('\n');
		comment.append(FXString(_stil->getAbsEntry(fileName.text())));
		FXString bug=FXString(_stil->getAbsBug(fileName.text()));
		if (!bug.empty()) {
			comment.append('\n');
			comment.append(bug);
		}
		if (!comment.empty()) {
			const FX88591Codec codec;
			_metaData.addInfo(metaData::COMMENT, codec.mb2utf(comment.text()).text());
		}
	}

	songLoaded(fileName.text());
	return true;
}

void gmeBackend::getInfo() {

	track_info_t ti;
	const char* err=_emu->track_info(&ti);
	if (err) {
		fxmessage("Warning: %s\n", err);
		return;
	}

	_metaData.addInfo(metaData::TITLE, ti.song);
	_metaData.addInfo(metaData::ARTIST, ti.author);
	_metaData.addInfo(gettext("copyright"), ti.copyright);
	_metaData.addInfo(gettext("system"), ti.system);
	_metaData.addInfo(gettext("game"), ti.game);
	_metaData.addInfo(gettext("dumper"), ti.dumper);
	if (!_hasStilInfo)
		_metaData.addInfo(metaData::COMMENT, ti.comment);

	time(ti.length/1000);
}

bool gmeBackend::check(const char* error) {

	if (error) {
		delPtr(_emu);
		fxmessage("Error: %s\n", error);
		return false;
	}
	return true;
}

void gmeBackend::close() {

	delPtr(_emu);

	songLoaded(0);
}

bool gmeBackend::rewind() {

	if (_emu) {
		_emu->start_track(_emu->current_track());
		return true;
	}
	return false;
}

bool gmeBackend::subtune(const unsigned int i) {

	if (_emu && (i>0) && (i<=_emu->track_count())) {
		_emu->start_track(i-1);
		getInfo();
		return true;
	}
	return false;
}

void gmeBackend::openAsma(const FXString& asmaPath) {

	if (asmaPath.empty())
		return;

	if (!_stil)
		_stil=new STIL(ASMA_STIL, ASMA_BUGLIST);

	if (!_stil->setBaseDir(asmaPath.text())) {
		fxmessage("%s\n", _stil->getErrorStr());
		delPtr(_stil);
	}
}

/*****************************************************************/

#include "iconFactory.h"

#define SETTINGS	gmeBackend::_settings

FXDEFMAP(gmeConfig) gmeConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	gmeConfig::ID_SAMPLERATE,	gmeConfig::onCmdSamplerate),
	FXMAPFUNC(SEL_COMMAND,	gmeConfig::ID_ASMA,		gmeConfig::onCmdAsma),
};

FXIMPLEMENT(gmeConfig, configFrame, gmeConfigMap, ARRAYNUMBER(gmeConfigMap))

gmeConfig::gmeConfig(FXPacker* win) :
	configFrame(win, gmeBackend::name, CREDITS, LINK) {

	FXint val;

	_asmaDir.connect(SETTINGS.asmaPath);

	new FXLabel(matrix(), tr("Samplerate"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, gmeConfig::ID_SAMPLERATE, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (SETTINGS.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);

	FXHorizontalFrame *hf=new FXHorizontalFrame(extraBottom());
	new FXLabel(hf, tr("ASMA path:"));
	new FXTextField(hf, 40, &_asmaDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(hf, tr("&Browse\tSelect ASMA directory"), GET_ICON(icon_documentopen), this, gmeConfig::ID_ASMA, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);
}

long gmeConfig::onCmdSamplerate(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplerate=11025;
		break;
	case 1:
		SETTINGS.samplerate=22050;
		break;
	case 2:
		SETTINGS.samplerate=44100;
		break;
	case 3:
		SETTINGS.samplerate=48000;
		break;
	}
	return 1;
}

long gmeConfig::onCmdAsma(FXObject*, FXSelector, void*) {

	FXDirDialog dirDialog(this, tr("Select ASMA directory"));
	dirDialog.setDirectory(SETTINGS.asmaPath);
	if (dirDialog.execute())
		SETTINGS.asmaPath=dirDialog.getDirectory();

	return 1;
}

#endif
