/*
 *  Copyright (C) 2007-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "hvlBackend.h"

#ifdef BUILD_HVL

#include "utils.h"

#define EXT	"*.(ahx|hvl),(ahx|hvl).*"

#define CREDITS		"HivelyTracker\nDexter/Abyss, Xeron/IRIS, and Pieknyman"
#define LINK		"http://code.google.com/p/hivelytracker/"

const char hvlBackend::name[]="Hively";

hvlConfig_t hvlBackend::_settings;

/*****************************************************************/

size_t hvlBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	size_t offset=0;

	if (_left) {
		memcpy(buffer, _buffer+_size-_left, _left);
		offset=_left;
		_left=0;
	}

	do {
		hvl_DecodeFrame(_tune, ((int8*)buffer)+offset, ((int8*)buffer)+offset+2, 4);
		offset+=_size;
	} while (offset<bufferSize-_size);

	size_t left=bufferSize-offset;
	if (left) {
		hvl_DecodeFrame(_tune, (int8*)_buffer, ((int8*)_buffer)+2, 4);
		memcpy((char*)buffer+offset, _buffer, left);
		_left=_size-left;
	}

	return bufferSize;
}

/*****************************************************************/

bool hvlBackend::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString hvlBackend::ext() const { return FXString(EXT); }

hvlBackend::hvlBackend() :
	inputBackend(name),
	_tune(0),
	_buffer(0) {

	loadSettings();

	hvl_InitReplayer();
}

hvlBackend::~hvlBackend() {

	close();
}

void hvlBackend::loadSettings() {

	_settings.samplerate=load("Samplerate", 44100);
}

void hvlBackend::saveSettings() {

	save("Samplerate", _settings.samplerate);
}

bool hvlBackend::open(const FXString& fileName) {

	close();

	_tune=hvl_LoadTune((TEXT*)fileName.text(), _settings.samplerate, 4);

	if (!_tune)
		return false;

	hvl_InitSubsong(_tune, 0);

	_metaData.addInfo(metaData::TITLE, (char*)_tune->ht_Name);
	FXString comment=FXString::null;
	for (unsigned int i=0; i<_tune->ht_InstrumentNr; i++)
		comment.append(FXStringFormat("%s\n", _tune->ht_Instruments[i].ins_Name));

	comment.trim();
	_metaData.addInfo(metaData::COMMENT, comment.text());

	_left=0;
	_size=(_settings.samplerate*4)/50;
	_buffer=new char[_size];

	songLoaded(fileName.text());
	return true;
}

void hvlBackend::close() {

	if (_tune) {
		hvl_FreeTune(_tune);
		delete[] _buffer;
		_buffer=0;
	}

	songLoaded(0);
}

bool hvlBackend::rewind() {

	if (_tune) {
		hvl_InitSubsong(_tune, _tune->ht_SongNum);
		return true;
	}
	return false;
}

bool hvlBackend::subtune(const unsigned int i) {

	if (_tune) {
		hvl_InitSubsong(_tune, i);
		return true;
	}
	return false;
}

/*****************************************************************/

#define SETTINGS	hvlBackend::_settings

FXDEFMAP(hvlConfig) hvlConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	hvlConfig::ID_SAMPLERATE,	hvlConfig::onCmdSamplerate),
};

FXIMPLEMENT(hvlConfig, configFrame, hvlConfigMap, ARRAYNUMBER(hvlConfigMap))

hvlConfig::hvlConfig(FXPacker* win) :
	configFrame(win, hvlBackend::name, CREDITS, LINK) {

	FXint val;

	new FXLabel(matrix(), tr("Samplerate"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, hvlConfig::ID_SAMPLERATE, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (SETTINGS.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);
}

long hvlConfig::onCmdSamplerate(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplerate=11025;
		break;
	case 1:
		SETTINGS.samplerate=22050;
		break;
	case 2:
		SETTINGS.samplerate=44100;
		break;
	case 3:
		SETTINGS.samplerate=48000;
		break;
	}
	return 1;
}

#endif
