/*
 *  Copyright (C) 2006-2011 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "madBackend.h"

#ifdef HAVE_MAD_H

#include "utils.h"
#include "tag.h"
#include "loadsym.h"

// mp3|mp2
#define EXT	"*.mp3"

#define CREDITS		FXStringFormat("%s\n%s", madBackend::dl_mad_version, madBackend::dl_mad_copyright).text()
#define LINK		"http://www.underbit.com/products/mad/"

const char madBackend::name[]="Mad";

#ifdef _WIN32
# define MADLIB	"libmad.dll"
#else
# define MADLIB	"libmad.so"
#endif

const AutoDLL madBackend::_dll(MADLIB);

const char* madBackend::dl_mad_version=0;
const char* madBackend::dl_mad_copyright=0;

void (*madBackend::dl_mad_stream_init)(struct mad_stream*)=0;
void (*madBackend::dl_mad_stream_finish)(struct mad_stream*)=0;
void (*madBackend::dl_mad_stream_buffer)(struct mad_stream*, unsigned char const*, unsigned long)=0;
int (*madBackend::dl_mad_stream_sync)(struct mad_stream*)=0;

void (*madBackend::dl_mad_header_init)(struct mad_header*)=0;
int (*madBackend::dl_mad_header_decode)(struct mad_header*, struct mad_stream*)=0;

void (*madBackend::dl_mad_frame_init)(struct mad_frame*)=0;
void (*madBackend::dl_mad_frame_finish)(struct mad_frame*)=0;
int (*madBackend::dl_mad_frame_decode)(struct mad_frame*, struct mad_stream*)=0;
void (*madBackend::dl_mad_frame_mute)(struct mad_frame*)=0;

void (*madBackend::dl_mad_synth_init)(struct mad_synth*)=0;
void (*madBackend::dl_mad_synth_mute)(struct mad_synth*)=0;
void (*madBackend::dl_mad_synth_frame)(struct mad_synth*, struct mad_frame const*)=0;

madConfig_t madBackend::_settings;

/*****************************************************************/

#if defined(HAVE_MMAP) || defined(_WIN32)
size_t madBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	size_t n=0;
	const size_t bufSize=bufferSize/sizeof(FXint);
	FXint* const buf=(FXint*)buffer;
	do {
		if (!_idx) {
			int res;
			do {
				res=dl_mad_frame_decode(&_frame, &_stream);
				if (res && !MAD_RECOVERABLE(_stream.error))
					return n;
			} while (res);

			dl_mad_synth_frame(&_synth, &_frame);
		}

		do {
			for (unsigned int j=0; j<_channels; j++)
				buf[n++]=_synth.pcm.samples[j][_idx];
			_idx++;
		} while ((n<bufSize) && (_idx<_synth.pcm.length));

		if (_idx>=_synth.pcm.length)
			_idx=0;
	} while (n<bufSize);

	return n;
}
#else
//TODO
#endif

/*****************************************************************/

bool madBackend::init() {

	if (!_dll.loaded())
		return false;

	LOADSYM(_dll, mad_version, const char*)
	LOADSYM(_dll, mad_copyright, const char*)
	LOADSYM(_dll, mad_stream_init, void(*)(mad_stream*))
	LOADSYM(_dll, mad_stream_finish, void(*)(mad_stream*))
	LOADSYM(_dll, mad_stream_buffer, void(*)(mad_stream*, unsigned char const*, unsigned long))
	LOADSYM(_dll, mad_stream_sync, int(*)(mad_stream*))
	LOADSYM(_dll, mad_header_init, void(*)(mad_header*))
	LOADSYM(_dll, mad_header_decode, int(*)(mad_header*, mad_stream*))
	LOADSYM(_dll, mad_frame_init, void(*)(mad_frame*))
	LOADSYM(_dll, mad_frame_finish, void(*)(mad_frame*))
	LOADSYM(_dll, mad_frame_decode, int(*)(mad_frame*, mad_stream*))
	LOADSYM(_dll, mad_frame_mute, void(*)(mad_frame*))
	LOADSYM(_dll, mad_synth_init, void(*)(mad_synth*))
	LOADSYM(_dll, mad_synth_mute, void(*)(mad_synth*))
	LOADSYM(_dll, mad_synth_frame, void(*)(mad_synth*, mad_frame const*))

	return true;
}

bool madBackend::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString madBackend::ext() const { return FXString(EXT); }

madBackend::madBackend() :
	inputBackend(name)
	{

	loadSettings();
}

madBackend::~madBackend() {

	close();
}

void madBackend::loadSettings() {

	_settings.fastscan=load("Fastscan", FALSE);
}

void madBackend::saveSettings() {

	save("Fastscan", _settings.fastscan);
}

#if defined(HAVE_MMAP) || defined(_WIN32)
bool madBackend::open(const FXString& fileName) {

	close();

	if (!_memMap.mapFile(fileName))
		return false;

	const tag tagPtr(&_memMap);
	FXTRACE((10, "offsBegin: %d.\n", tagPtr.offsBegin()));
	FXTRACE((10, "offsEnd: %d.\n", tagPtr.offsEnd()));

	_data=(FXuchar*)_memMap.base()+tagPtr.offsBegin();
	_length=_memMap.length()-(tagPtr.offsBegin()+tagPtr.offsEnd());

	dl_mad_stream_init(&_stream);
	mad_stream_options(&_stream, MAD_OPTION_IGNORECRC);

	int res;

	dl_mad_stream_buffer(&_stream, _data, _length);
	if (dl_mad_stream_sync(&_stream)<0)
		goto error;

	mad_header header;
	dl_mad_header_init(&header);
	do {
		res=dl_mad_header_decode(&header, &_stream);
	} while (res);

	{
	const unsigned int offset=((header.flags&MAD_FLAG_LSF_EXT)?(header.mode?21:13):(header.mode?36:21));
	const int xingFrames=getXing((char*)_stream.buffer+offset);

	if (xingFrames) {
		do {
			res=dl_mad_header_decode(&header, &_stream);
		} while (res);
	}

	_samplerate=header.samplerate;
	if (!_samplerate)
		goto error;

	unsigned int samples=32*MAD_NSBSAMPLES(&header);
	_channels=MAD_NCHANNELS(&header);

	if (xingFrames)
		time(xingFrames*samples/_samplerate);
	else {
		bool cbr=true;
		int count=0;
		const unsigned int bitrate=header.bitrate;
		do {
			res=dl_mad_header_decode(&header, &_stream);
			if (!res) {
				samples+=32*MAD_NSBSAMPLES(&header);
				if (_samplerate!=header.samplerate) {
					fxmessage("Samplerate change not supported\n");
					mad_header_finish(&header);
					goto error;
				}
				if (cbr && (bitrate!=header.bitrate)) {
					FXTRACE((10, "VBR\n"));
					cbr=false;
				}
				if ((++count>10) && (cbr || _settings.fastscan)) {
					time(_length/(bitrate>>3));
					break;
				}
			}
		} while ((!res) || MAD_RECOVERABLE(_stream.error));

		if ((!cbr) && (!_settings.fastscan))
			time(samples/_samplerate);
	}
	}

	mad_header_finish(&header);

	dl_mad_stream_buffer(&_stream, _data, _length);
	dl_mad_stream_sync(&_stream);

	dl_mad_frame_init(&_frame);
	dl_mad_synth_init(&_synth);

	_metaData.addInfo(metaData::TITLE, tagPtr.title());
	_metaData.addInfo(metaData::ARTIST, tagPtr.artist());
	_metaData.addInfo(metaData::ALBUM, tagPtr.album());
	_metaData.addInfo(metaData::GENRE, tagPtr.genre());
	_metaData.addInfo(metaData::YEAR, tagPtr.year());
	_metaData.addInfo("publisher", tagPtr.publisher());
	_metaData.addInfo(metaData::TRACK, tagPtr.track());
	_metaData.addInfo(metaData::COMMENT, tagPtr.comment());
	_metaData.addInfo(tagPtr.image());

	_idx=0;

	songLoaded(fileName.text());
	return true;

error:
	dl_mad_stream_finish(&_stream);
	_memMap.unmap();

	return false;
}

int madBackend::getXing(const char* buf) {

	if ((!tag::isFrame(buf, "Xing")) &&
	(!tag::isFrame(buf, "Info")))
		return 0;

	FXTRACE((10, "Xing tag found.\n"));

	if (!buf[7]&0x01)
		// No Frames field
		return 0;

	const int frames=tag::getFrameSize(buf+8, false);
	FXTRACE((10, "Frames: %d.\n", frames));

	return frames;
}

void madBackend::close() {

	if (!songLoaded())
		return;

	dl_mad_frame_finish(&_frame);
	mad_synth_finish(&_synth);
	dl_mad_stream_finish(&_stream);

	_memMap.unmap();

	songLoaded(0);
}

bool madBackend::rewind() {

	if (songLoaded()) {
		dl_mad_stream_buffer(&_stream, _data, _length);
		dl_mad_stream_sync(&_stream);

		dl_mad_frame_mute(&_frame);
		dl_mad_synth_mute(&_synth);

		_idx=0;

		return true;
	}
	return false;
}
#else
//TODO
#endif

/*****************************************************************/

#define SETTINGS	madBackend::_settings

FXIMPLEMENT(madConfig, configFrame, 0, 0)

madConfig::madConfig(FXPacker* win) :
	configFrame(win, madBackend::name, CREDITS, LINK) {

	_scan.connect(SETTINGS.fastscan);

	FXVerticalFrame *vert=new FXVerticalFrame(matrix());
	new FXCheckButton(vert, tr("Fast scan\tScan only few frame of file, time detection may be inaccurate"), &_scan, FXDataTarget::ID_VALUE);
}

#endif
