/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "oggBackend.h"

#ifdef HAVE_VORBIS

#include "utils.h"

#include <FXUTF8Codec.h>

ov_callbacks oggBackend::vorbis_callbacks = {
    oggBackend::read_func,
    oggBackend::seek_func,
    oggBackend::close_func,
    oggBackend::tell_func
};

#define EXT	"*.ogg"

#define CREDITS		"OggVorbis\nCopyright \302\251 Xiph.org Foundation"
#define LINK		"http://www.xiph.org/"

const char oggBackend::name[]="Ogg-Vorbis";

oggConfig_t oggBackend::_settings;

/*****************************************************************/

size_t oggBackend::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

	size_t n=0;
	long read;
	do {
		int current_section;
		read=ov_read(_vf, (char*)buffer+n, bufferSize-n, 0,
			(_settings.precision==S16)?2:1, (_settings.precision!=U8), &current_section);
		n+=read;
	} while (read && n<bufferSize);

	return n;
}

/*****************************************************************/

bool oggBackend::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString oggBackend::ext() const { return FXString(EXT); }

oggBackend::oggBackend() :
	inputBackend(name),
	_vf(0),
	_vi(0) {

	loadSettings();
}

oggBackend::~oggBackend() {

	close();
}

void oggBackend::loadSettings() {

	_settings.precision=(load("Bits", 16)==16)?S16:U8;
}

void oggBackend::saveSettings() {

	save("Bits", (_settings.precision==S16)?16:8);
}

bool oggBackend::open(const FXString& fileName) {

	close();

	if (!_file.open(fileName))
		return false;

	_vf=new OggVorbis_File;
	if (ov_open_callbacks(&_file, _vf, NULL, 0, vorbis_callbacks)<0) {
		delPtr(_vf);
		_file.close();
		return false;
	}

	_vi=ov_info(_vf, -1);

	time((FXint)ov_time_total(_vf, -1));

	FXString title=FXString::null;
	FXString artist=FXString::null;
	FXString year=FXString::null;
	FXString album=FXString::null;
	FXString genre=FXString::null;
	FXString comment=FXString::null;

	char **ptr=ov_comment(_vf,-1)->user_comments;
	while (*ptr) {
		FXTRACE((10, "%s\n", *ptr));
		if (!getComment(*ptr, &title, "title"))
		if (!getComment(*ptr, &artist, "artist"))
		if (!getComment(*ptr, &year, "date"))
		if (!getComment(*ptr, &album, "album"))
		if (!getComment(*ptr, &genre, "genre"))
		if (!getComment(*ptr, &comment, "comment")) {
			if (!comparecase(*ptr, "tracknumber", 11))
				_metaData.addInfo(metaData::TRACK, FXString(*ptr).erase(0, 12).text());
		}
		++ptr;
	}

//METADATA_BLOCK_PICTURE
/*
<32> 	 The picture type according to the ID3v2 APIC frame:
<32> 	 The length of the MIME type string in bytes.
<n*8> 	 The MIME type string
<32> 	 The length of the description string in bytes.
<n*8> 	 The description of the picture, in UTF-8. 
 <32> 	 The width of the picture in pixels. 
 <32> 	 The height of the picture in pixels. 
 <32> 	 The color depth of the picture in bits-per-pixel. 
 <32> 	 For indexed-color pictures (e.g. GIF), the number of colors used, or 0 for non-indexed pictures. 
 <32> 	 The length of the picture data in bytes. 
 <n*8> 	 The binary picture data.
*/

//deprecated:
//COVERARTMIME
//COVERART (base64 encoded)

	_metaData.addInfo(metaData::TITLE, title.text());
	_metaData.addInfo(metaData::ARTIST, artist.text());
	_metaData.addInfo(metaData::ALBUM, album.text());
	_metaData.addInfo(metaData::GENRE, genre.text());
	_metaData.addInfo(metaData::YEAR, year.text());
	_metaData.addInfo(metaData::COMMENT, comment.text());

	songLoaded(fileName.text());
	return true;
}

bool oggBackend::getComment(const char* orig, FXString* dest, const char* type) {

	const int n=strlen(type);
	if (comparecase(orig, type, n))
		return false;

	if (!dest->empty())
		dest->append(", ");

	const FXUTF8Codec codec;
	dest->append(codec.mb2utf(orig+n+1));
	return true;
}

void oggBackend::close() {

	if (_vf) {
		ov_clear(_vf);
		delPtr(_vf);
		_file.close();
	}

	_vi=0;

	songLoaded(0);
}

bool oggBackend::rewind() {

	if (!_vf)
		return false;

	if (ov_raw_seek(_vf, 0))
		return false;

	return true;
}

size_t oggBackend::read_func(void *ptr, size_t size, size_t nmemb, void *datasource) {

	return ((FXFile*)datasource)->readBlock((char*)ptr, size*nmemb);
}

int oggBackend::seek_func(void *datasource, ogg_int64_t offset, int whence) {

	FXuint from;
	switch (whence) {
	case SEEK_SET:
		from=FXIO::Begin;
		break;
	case SEEK_CUR:
		from=FXIO::Current;
		break;
	case SEEK_END:
		from=FXIO::End;
		break;
	default:
		return -1;
	}
	return ((FXFile*)datasource)->position(offset, from);
}

int oggBackend::close_func(void *datasource) {

	((FXFile*)datasource)->close();
	return 0;
}

long oggBackend::tell_func(void *datasource) {

	return ((FXFile*)datasource)->position();
}

/*****************************************************************/

#define SETTINGS	oggBackend::_settings

FXDEFMAP(oggConfig) oggConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	oggConfig::ID_BITS,		oggConfig::onCmdBits),
};

FXIMPLEMENT(oggConfig, configFrame, oggConfigMap, ARRAYNUMBER(oggConfigMap))

oggConfig::oggConfig(FXPacker* win) :
	configFrame(win, oggBackend::name, CREDITS, LINK) {

	FXint val;

	new FXLabel(matrix(), tr("Bits"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox* _bitBox=new FXListBox(matrix(), this, oggConfig::ID_BITS, LAYOUT_FILL|FRAME_NORMAL);
	_bitBox->fillItems("8\n16");
	_bitBox->setNumVisible(2);
	switch (SETTINGS.precision) {
	case U8:
		val=0;
		break;
	default:
	case S16:
		val=1;
		break;
	}
	_bitBox->setCurrentItem(val);
}

long oggConfig::onCmdBits(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.precision=U8;
		break;
	case 1:
		SETTINGS.precision=S16;
		break;
	}
	return 1;
}

#endif
