/*
 *  Copyright (C) 2006-2011 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "infoDialog.h"
#include "textLabel.h"
#include "settings.h"
#include "iconFactory.h"
#include "utils.h"
#include "metaData.h"
#include "imageData.h"

#if FOX_LEVEL<32
#  include <FXJPGImage.h>
#endif

#define IMAGESIZE 150

FXint imageLoader::run() {

	FXIconSource iconsource(FXApp::instance());
	FXImage *img;

#if FOX_LEVEL<32
	// Hack to overcome a FOX bug
	if (FXJPGImage::supported && !comparecase(FXPath::extension(_name), "jpeg"))
		img=iconsource.loadScaledImageFile(_name, IMAGESIZE, 0, FXJPGImage::fileExt);
	else
#endif
	img=iconsource.loadScaledImageFile(_name, IMAGESIZE, SETTINGS->albumArtQuality());

	_loaded->setData((void*)(FXival)img);
	_loaded->signal();

	return 0;
}

FXbool imageLoader::load(const FXString& file) {

	_name=file;

	return start();
}

/*****************************************************************/

FXDEFMAP(infoDialog) infoDialogMap[] = {
	// Message_Type			ID				Message_Handler
	FXMAPFUNC(SEL_IO_READ,		infoDialog::ID_IMG_LOADED,	infoDialog::onCmdImgLoaded),
};

FXIMPLEMENT(infoDialog, FXDialogBox, infoDialogMap, ARRAYNUMBER(infoDialogMap))

infoDialog::infoDialog(FXWindow* w) :
	FXDialogBox(w, FXString::null, DECOR_TITLE|DECOR_BORDER),
	_font(0),
	_imageLoader(0),
	_imgFrame(0) {

	new FXButton(this, tr("&Close"), GET_ICON(icon_dialogcancel), this, FXDialogBox::ID_CANCEL,
		BUTTON_INITIAL|BUTTON_DEFAULT|BUTTON_NORMAL|LAYOUT_CENTER_X|LAYOUT_SIDE_BOTTOM);
	new FXSeparator(this, SEPARATOR_GROOVE|LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM);


	FXFontDesc fontdesc;
	getApp()->getNormalFont()->getFontDesc(fontdesc);
	fontdesc.slant=FXFont::Italic;
	_italicFont=new FXFont(getApp(), fontdesc);

	FXHorizontalFrame *container=new FXHorizontalFrame(this, LAYOUT_FILL_X, 0,0,0,0, 0,0,0,0, 0,0);

	_imgFrame=new FXImageFrame(container, 0, LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|FRAME_SUNKEN|FRAME_THICK, 0, 0, IMAGESIZE, IMAGESIZE);

	_matrix=new FXMatrix(container, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL_X);

	_font=new FXFont(getApp(), "courier", 10,
		FXFont::Normal, FXFont::Straight,
		FONTENCODING_DEFAULT, FXFont::NonExpanded, FXFont::Fixed);
	_textFrame=new FXHorizontalFrame(this, LAYOUT_FILL|FRAME_NORMAL, 0,0,0,0, 0,0,0,0, 0,0);
	_text=new FXText(_textFrame, 0, 0, TEXT_READONLY|TEXT_WORDWRAP|LAYOUT_FILL);
	_text->setFont(_font);
}

void infoDialog::setInfo(const metaData* i) {

	_text->setText("");
	_text->setVisibleRows(0);
	_text->setVisibleColumns(0);

	FXString folder=i->getInfo(metaData::LOCATION);
	if (folder.empty()) {
		new FXLabel(_matrix, tr("No song loaded"));
		return;
	}

	FXWindow *w=_matrix->getFirst();
	while (w!=0) {
		FXWindow *n=w->getNext();
		delete w;
		w=n;
	}

	FXColor lblColor=altColor(getApp()->getForeColor());

	int j=-1;
	while ((j=i->moreInfo(j))>=0) {
		FXString temp=i->getInfo(j);
		if (temp.empty())
			continue;

		unsigned int rows=1;
		unsigned int cols=0;
		unsigned int cnt=0;
		FXuint pos=0;
		const FXint len=temp.length();
		const FXchar* s=temp.text();
		while (pos<len) {
			// Convert Mac and Windows EOLs to Unix
			if (s[pos]=='\r')
				temp.replace(pos, (s[pos+1]=='\n')?'\0':'\n');
			//Count rows and columns
			if (s[pos]=='\n') {
				rows++;
				if (cnt>cols)
					cols=cnt;
				cnt=0;
			} else
				cnt++;
			pos+=FXString::utfBytes[(FXuchar)s[pos]];
		}
		if (cnt>cols)
			cols=cnt;

		FXTRACE((10, "Rows: %d\n", rows));
		FXTRACE((10, "Cols: %d\n", cols));

		if (rows>1) {
			_text->setText(temp);
			_text->setVisibleRows((rows<20)?rows:20);
			_text->setVisibleColumns((cols<80)?cols+2:80);
		} else {
			const FXString info=i->getKey(j);
			if (compare(info, "location")) { // FIXME
				//info.append(':');
				FXLabel *lbl=new FXLabel(_matrix, tr(info.text()), 0, LAYOUT_FILL_X|JUSTIFY_RIGHT);
				lbl->setTextColor(lblColor);
				lbl->setFont(_italicFont);

				new textLabel(_matrix, temp, JUSTIFY_LEFT, 32);
			}
		}
	}

	// Display album art
	FXImage* newimg=0;
	if (imageData *img=i->getImage()) {
		FXMemoryStream stream;
		if (stream.open(FXStreamLoad, img->size(), img->data())) {
			FXIconSource iconsource(FXApp::instance());
			newimg=iconsource.loadScaledImageStream(stream, IMAGESIZE,
				SETTINGS->albumArtQuality(), img->type().substitute("image/", ""));
			stream.close();
		}
	}

	if (newimg) {
		FXTRACE((10, "Using embedded cover art\n"));
		setNewImage(newimg);
	} else {
		FXGUISignal* imgLoaded=new FXGUISignal(getApp(), this, infoDialog::ID_IMG_LOADED);
		_imageLoader=new imageLoader(imgLoaded);

		FXString *_files;
		folder=FXPath::directory(folder);
		FXint n=FXDir::listFiles(_files, folder, "(cover|front|folder).*", FXDir::CaseFold);

		if (n) {
			FXString file=FXStringFormat("%s%c%s", folder.text(), PATHSEP, _files[0].text());

			FXTRACE((10, "Trying to load album art: %s\n", file.text()));
			_imageLoader->load(file);
			getApp()->beginWaitCursor();
		} else {
			_imgFrame->hide();
		}

		delete [] _files;
	}

	if (!_matrix->getFirst())
		new FXLabel(_matrix, tr("No info"), 0, LAYOUT_FILL_X);
	if (_text->getLength()) {
		_textFrame->show();
		_text->resize(_text->getDefaultWidth(), _text->getDefaultHeight());
	} else {
		_textFrame->hide();
	}
	this->create();
	this->resize(this->getDefaultWidth(), this->getDefaultHeight());
}

infoDialog::~infoDialog() {

	if (_imageLoader) {
		_imageLoader->join();
		delete _imageLoader;
		getApp()->endWaitCursor();
	}

	setNewImage(0);

	delete _font;
	delete _italicFont;
}

void infoDialog::setNewImage(FXImage* newImage) {

	FXImage *img=_imgFrame->getImage();
	_imgFrame->setImage(newImage);
	delete img;
}

long infoDialog::onCmdImgLoaded(FXObject*, FXSelector, void* ptr) {

	_imageLoader->join();

	setNewImage(0);

	if (ptr) {
		FXImage *img=(FXImage*)ptr;
		img->create();
		_imgFrame->setImage(img);
	} else {
		_imgFrame->hide();
	}

	getApp()->endWaitCursor();
	delPtr(_imageLoader);

	return 1;
}
