/*
 *  Copyright (C) 2006-2008 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "playlist.h"
#include "utils.h"
#include "iconFactory.h"
#include "trackList/trackListFactory.h"

FXuint playlist::_seed=(FXuint)FXSystem::now();

FXDEFMAP(playlist) playlistMap[]={

	//FXMAPFUNC(SEL_DRAGGED,			0,	playlist::onDragged),
	FXMAPFUNC(SEL_DND_MOTION,		0,	playlist::onDNDMotion),
	FXMAPFUNC(SEL_DND_DROP,			0,	playlist::onDNDDrop),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	0,	playlist::onRgtClk),
	FXMAPFUNC(SEL_COMMAND,		playlist::ID_ITEM_DEL,			playlist::onCmdDel),
	FXMAPFUNCS(SEL_UPDATE,	playlist::ID_SORT_ASC,playlist::ID_SORT_RANDOM,	playlist::onUpdSort),
	FXMAPFUNCS(SEL_COMMAND,	playlist::ID_SORT_ASC,playlist::ID_SORT_RANDOM,	playlist::onCmdSort),
};

FXIMPLEMENT(playlist, list, playlistMap, ARRAYNUMBER(playlistMap))

bool playlist::add(const FXString& item) {

	if (findItem(item)==-1) {
		FXTRACE((10, "Add %s to playlist\n", item.text()));
		const FXint i=appendItem(item);
		if (!i)
			setCurrentItem(i, TRUE);
		return true;
	}
	FXTRACE((10, "Duplicate item\n"));
	return false;
}

long playlist::onDNDDrop(FXObject*, FXSelector, void*) {

	FXString item;
	getDNDData(FROM_DRAGNDROP, urilistType, item);
	dropFinished(DRAG_COPY);

	if (add(FXURL::fileFromURL(item)))
		sortItems();

	return 1;
}

long playlist::onDNDMotion(FXObject*, FXSelector, void* ptr) {

	FXString item;
	acceptDrop(getDNDData(FROM_DRAGNDROP, urilistType, item) && FXStat::isFile(FXURL::fileFromURL(item))?
		DRAG_COPY:DRAG_REJECT);

	return 1;
}

long playlist::onRgtClk(FXObject* sender, FXSelector sel, void* ptr) {

	list::onRightBtnRelease(sender, sel, ptr);
	FXEvent *event=(FXEvent*)ptr;
	if (event->moved)
		return 1;

	FXint item=getItemAt(event->win_x, event->win_y);
	setUserData((void*)(FXival)item);

	FXMenuPane pane(this);
	if (item>=0) {
		new FXLabel(&pane, shrink(getItemText(item)));
		new FXMenuSeparator(&pane);
		new FXMenuCommand(&pane, tr("Remove item\t\tRemove selected item from playlist"),
			GET_ICON(icon_listremove), this, playlist::ID_ITEM_DEL);
		new FXMenuSeparator(&pane);
	}
	new FXMenuRadio(&pane, tr("Sort ascending\t\tSort ascending"), this, playlist::ID_SORT_ASC);
	new FXMenuRadio(&pane, tr("Sort descending\t\tSort descending"), this, playlist::ID_SORT_DESC);
	new FXMenuRadio(&pane, tr("Shuffle\t\tSort randomly"), this, playlist::ID_SORT_RANDOM);
	new FXMenuSeparator(&pane);
	new FXMenuCommand(&pane, tr("Cancel\t\tCancel action"), GET_ICON(icon_actioncancel));
	pane.create();
	pane.popup(0, event->root_x, event->root_y);
	getApp()->runModalWhileShown(&pane);
	return 1;
}

long playlist::onCmdDel(FXObject*, FXSelector, void*) {

	const FXint idx=(FXint)(FXival)getUserData();
	if (idx<0)
		return 0;

	removeItem(idx);

	return 1;
}

long playlist::onCmdSort(FXObject*, FXSelector sel, void*) {

	switch (FXSELID(sel)) {
	case ID_SORT_ASC:
		setSortFunc(playlist::ascending);
		break;
	case ID_SORT_DESC:
		setSortFunc(playlist::descending);
		break;
	case ID_SORT_RANDOM:
		setSortFunc(playlist::random);
		break;
	}

	if (!getNumItems())
		return 0;

	sortItems();
	makeItemVisible(getCurrentItem());

	return 1;
}

long playlist::onUpdSort(FXObject* sender, FXSelector sel, void*) {

	bool check=false;
	switch (FXSELID(sel)) {
	case ID_SORT_ASC:
		check=(getSortFunc()==playlist::ascending);
		break;
	case ID_SORT_DESC:
		check=(getSortFunc()==playlist::descending);
		break;
	case ID_SORT_RANDOM:
		check=(getSortFunc()==playlist::random);
		break;
	}
	sender->handle(this, check?FXSEL(SEL_COMMAND, ID_CHECK):FXSEL(SEL_COMMAND, ID_UNCHECK), 0);
	return 1;
}

FXint playlist::ascending(const FXListItem* a, const FXListItem* b) {

	return comparecase(FXPath::name(a->getText()), FXPath::name(b->getText()));
}

FXint playlist::descending(const FXListItem* a, const FXListItem* b) {

	return comparecase(FXPath::name(b->getText()), FXPath::name(a->getText()));
}

FXint playlist::random(const FXListItem*, const FXListItem*) {

	return fxrandom(_seed);
}

void playlist::sorting(const char* sort) {

	FXListSortFunc lsf=0;
	if (!compare(sort, "ascending"))
		lsf=playlist::ascending;
	else if (!compare(sort, "descending"))
		lsf=playlist::descending;
	else if (!compare(sort, "random"))
		lsf=playlist::random;

	setSortFunc(lsf);
}

const char* playlist::sorting() {

	if (getSortFunc()==playlist::ascending)
		return "ascending";
	else if (getSortFunc()==playlist::descending)
		return "descending";
	else if (getSortFunc()==playlist::random)
		return "random";
	return 0;
}

int playlist::load(const FXString& path) {

	delPtr(_tracks);

	trackList* tracklist=TFACTORY->get(path);

	if (!tracklist)
		return 0;

	_tracks=tracklist->load();

	delete tracklist;

	return _tracks->no();
}

bool playlist::save(const FXString& file) {

	trackList* tracklist=TFACTORY->get(file);

	if (!tracklist)
		return 0;

	//FIXME
	if (!_tracks)
		_tracks=new tracks();
	_tracks->clear();
	for (FXint i=0; i<getNumItems(); i++) {
		_tracks->append(getItemText(i));
	}

	const bool temp=tracklist->save(_tracks);

	delete tracklist;

	return temp;
}

int playlist::filter(const FXString& filter) {

	if (!_tracks)
		return 0;

	FXTRACE((10, "filter: %s\n", filter.text()));
	clearItems();

	for (FXint i=0; i<_tracks->no(); i++) {
		FXTRACE((10, "File: %s\n", _tracks->at(i).location().text()));
		if (FXPath::match(filter, FXPath::name(_tracks->at(i).location()),
			FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD) &&
			FXStat::exists(_tracks->at(i).location())) {
			appendItem(_tracks->at(i).location(), 0, (void*)&_tracks->at(i));
		}
	}

	return getNumItems();
}
