/*
 *  Copyright (C) 2007-2009 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "tag.h"

#include "genres.h"

#include <FX88591Codec.h>
#include <FXUTF16Codec.h>
#include <FXUTF8Codec.h>

#include "utils.h"

#define ID3V1_TAG_SIZE		128

#define ID3V2_HEADER_SIZE	10
#define ID3V2_EXT_HEADER_SIZE	6

#define APE_HEADER_SIZE		32

tag::tag(FXFile* file) :
	_title(FXString::null),
	_artist(FXString::null),
	_album(FXString::null),
	_publisher(FXString::null),
	_comment(FXString::null),
	_genre(FXString::null),
	_year(FXString::null),
	_track(FXString::null),
	_offsBegin(0),
	_offsEnd(0),
	_img(0) {

	char buf[ID3V1_TAG_SIZE];
	int id3v1Size=0;

	// Check for an APE tag at the end of file
	// it may be confused as an ID3v1 tag
	file->position(-APE_HEADER_SIZE, FXIO::End);
	file->readBlock((void*)buf, 8);
	if (!isFrame(buf, "APETAGEX")) {
		// Check for ID3v1 tag
		file->position(-ID3V1_TAG_SIZE, FXIO::End);
		file->readBlock((void*)buf, ID3V1_TAG_SIZE);
		id3v1Size=getID3v1(buf);
	}

	// Check for ID3v2 tag
	file->position(0);
	file->readBlock((void*)buf, ID3V2_HEADER_SIZE);

	int version=0;
	int tagSize=0;
	int extHdrSize=0;
	if (checkID3v2(buf, true)) {
		if (parseID3v2header(buf, version, tagSize)) {
			FXTRACE((10, "Extended header found\n"));
			file->readBlock((void*)buf, ID3V2_EXT_HEADER_SIZE);
			extHdrSize=getExtHdrSize(buf, version);
			if (extHdrSize)
				file->position(extHdrSize, FXIO::Current);
		}
	} else {
		file->position(-(ID3V2_HEADER_SIZE+id3v1Size), FXIO::End);
		file->readBlock((void*)buf, ID3V2_HEADER_SIZE);
		if (checkID3v2(buf, false)) {
			if (parseID3v2header(buf, version, tagSize)) {
				FXTRACE((10, "Extended header found\n"));
				file->readBlock((void*)buf, ID3V2_EXT_HEADER_SIZE);
				extHdrSize=getExtHdrSize(buf, version);
			}
			file->position(-(tagSize+ID3V2_EXT_HEADER_SIZE), FXIO::Current);
		}
	}

	int id3v2Size=0;
	if (tagSize) {
		id3v2Size=ID3V2_HEADER_SIZE+tagSize;
		if (extHdrSize)
			id3v2Size+=ID3V2_EXT_HEADER_SIZE+extHdrSize;

		char *dataBuf=new char[tagSize];
		file->readBlock((void*)dataBuf, tagSize);

		int i=0;
		while (i<tagSize && dataBuf[i])
			i+=(version==2)?getID3v2_2Frame(dataBuf+i)+6:getID3v2Frame(dataBuf+i, version)+10;

		delete[] dataBuf;
	}

	if (id3v2Size)
		return;

	// Check for APE tag
	file->position(0);
	file->readBlock((void*)buf, APE_HEADER_SIZE);

	int itemsSize=0;
	if (!checkAPE(buf, itemsSize, tagSize)) {
		file->position(-(APE_HEADER_SIZE+id3v1Size), FXIO::End);
		file->readBlock((void*)buf, APE_HEADER_SIZE);
		if (!checkAPE(buf, itemsSize, tagSize))
			return;
		file->position(-(APE_HEADER_SIZE+itemsSize), FXIO::Current);
	}

	char* dataBuf=new char[itemsSize];
	file->readBlock((void*)dataBuf, itemsSize);

	int i=0;
	while (i<itemsSize && dataBuf[i])
		i+=getAPEItem(dataBuf+i);

	delete[] dataBuf;
};

#if defined(HAVE_MMAP) || defined(_WIN32)
tag::tag(FXMemMap* map) :
	_title(FXString::null),
	_artist(FXString::null),
	_album(FXString::null),
	_publisher(FXString::null),
	_comment(FXString::null),
	_genre(FXString::null),
	_year(FXString::null),
	_track(FXString::null),
	_offsBegin(0),
	_offsEnd(0),
	_img(0) {

	int id3v1Size=0;

	// Check for an APE tag at the end of file
	// it may be confused as an ID3v1 tag
	char *buf=(char*)map->base()+map->length()-APE_HEADER_SIZE;
	if (!isFrame(buf, "APETAGEX")) {
		// Check for ID3v1 tag
		buf=(char*)map->base()+map->length()-ID3V1_TAG_SIZE;
		id3v1Size=getID3v1(buf);
		_offsEnd=id3v1Size;
	}

	// Check for ID3v2 tag
	int version=0;
	int tagSize=0;
	int extHdrSize=0;

	FXint* offset=0;

	buf=(char*)map->base();
	if (checkID3v2(buf, true)) {
		// ID3v2 tag at the beginning of file
		if (parseID3v2header(buf, version, tagSize)) {
			FXTRACE((10, "Extended header found\n"));
			extHdrSize=getExtHdrSize(buf, version)+ID3V2_EXT_HEADER_SIZE;
		}
		offset=&_offsBegin;
	} else {
		buf=(char*)map->base()+map->length()-(ID3V2_HEADER_SIZE+id3v1Size);
		if (checkID3v2(buf, false)) {
			// ID3v2 tag at the end of file
			if (parseID3v2header(buf, version, tagSize)) {
				FXTRACE((10, "Extended header found\n"));
				extHdrSize=getExtHdrSize(buf, version)+ID3V2_EXT_HEADER_SIZE;
			}
			buf-=tagSize+extHdrSize;
			offset=&_offsEnd;
		}
	}

	if (tagSize) {
		int id3v2Size=ID3V2_HEADER_SIZE+extHdrSize+tagSize;

		buf+=ID3V2_HEADER_SIZE+extHdrSize;
		*offset+=id3v2Size;

		int i=0;
		while (i<tagSize && buf[i])
			i+=(version==2)?getID3v2_2Frame(buf+i)+6:getID3v2Frame(buf+i, version)+10;

		return;
	}

	// Check for APE tag
	int itemsSize=0;

	buf=(char*)map->base();
	if (checkAPE(buf, itemsSize, tagSize)) {
		offset=&_offsBegin;
	} else {
		buf=(char*)map->base()+map->length()-(APE_HEADER_SIZE+id3v1Size);
		if (!checkAPE(buf, itemsSize, tagSize))
			return;
		offset=&_offsEnd;
	}

	buf+=APE_HEADER_SIZE;
	*offset+=tagSize;

	int i=0;
	while (i<itemsSize && buf[i])
		i+=getAPEItem(buf+i);
};
#endif

/************************/

bool tag::isFrame(const char* buf, const char* frame) {

	for (unsigned int i=0; i<strlen(frame); i++)
		if (buf[i]!=frame[i])
			return false;
	return true;
}

FXint tag::getFrameSize(const char* frame, bool synchsafe) {

	return synchsafe?
		(FXuint)((FXuchar)(frame[0]&0x7f))<<21|(FXuint)((FXuchar)(frame[1]&0x7f))<<14|
		(FXuint)((FXuchar)(frame[2]&0x7f))<<7|(FXuint)((FXuchar)(frame[3]&0x7f)):
		(FXuint)((FXuchar)frame[0])<<24|(FXuint)((FXuchar)frame[1])<<16|
		(FXuint)((FXuchar)frame[2])<<8|(FXuint)((FXuchar)frame[3]);
}

/******** ID3v1 ********/

FXString tag::getString(const char* ptr, const char max) {

	const FX88591Codec codec;
	unsigned char len=0;
	while (ptr[len] && (len<max))
		len++;
	return codec.mb2utf(ptr, len).simplify();
}

int tag::getID3v1(char* buf) {

	if (!isFrame(buf, "TAG"))
		return 0;

	FXTRACE((10, "ID3v1 tag found.\n"));

	if (_title.empty())
		_title=getString(buf+3);

	if (_artist.empty())
		_artist=getString(buf+33);

	if (_album.empty())
		_album=getString(buf+63);

	if (_year.empty()) {
		_year=getString(buf+93, 4);
		if (_year.length()<4)
			_year.length(0);
	}

	if (_comment.empty())
		_comment=getString(buf+97);

	unsigned char tmp=(unsigned char)buf[126];
	if (!_track && !buf[125] && tmp)
#if FOX_MINOR>6
		_track=FXString::value(tmp);
#else
		_track=FXStringVal(tmp);
#endif

	tmp=(unsigned char)buf[127];
	if (_genre.empty() && tmp<GENRES)
		_genre=FXString(::genre[tmp]);

	return ID3V1_TAG_SIZE;
}

/******** ID3v2 ********/

FXString tag::getID3v2Text(const char* buf, int size) {

	// text encoding
	// 00 - ISO-8859-1
	// 01 - UTF-16
	// 02 - UTF-16BE
	// 03 - UTF-8

	const char encoding=*buf++;
	FXTRACE((10, "ID3v2 Frame encoding: %d\n", encoding));

	FXTextCodec *codec=0;
	switch (encoding) {
	case 0:
		codec=new FX88591Codec();
		break;
	case 1:
		codec=new FXUTF16Codec();
		break;
	case 2:
		codec=new FXUTF16BECodec();
		break;
	case 3:
		codec=new FXUTF8Codec();
		break;
	default:
		return FXString::null;
	}

	FXString tmp=codec->mb2utf(buf, size);

	delete codec;

	// Trim null char at the end of string
	const FXint len=tmp.length()-1;
	if (!tmp.text()[len])
		tmp.length(len);

	tmp.substitute('\0', ',');

	return tmp;
}

int tag::getID3v2_2Frame(const char* buf) {

	const int size=(FXuint)((FXuchar)buf[3])<<16|(FXuint)((FXuchar)buf[4])<<8|(FXuint)((FXuchar)buf[5]);
	FXTRACE((10, "ID3v2 Frame: %.3s, size: %d\n", buf, size));

	if (isFrame(buf, "TT2")) {
		_title=getID3v2Text(buf+6, size-1);
		FXTRACE((10, "ID3v2 title: %s\n", _title.text()));
	} else
	if (isFrame(buf, "TP1")) {
		_artist=getID3v2Text(buf+6, size-1);
		FXTRACE((10, "ID3v2 artist: %s\n", _artist.text()));
	} else
	if (isFrame(buf, "TAL")) {
		_album=getID3v2Text(buf+6, size-1);
		FXTRACE((10, "ID3v2 album: %s\n", _album.text()));
	} else
	if (isFrame(buf, "TRK")) {
		_track=getID3v2Text(buf+6, size-1).before('/');
		FXTRACE((10, "ID3v2 track: %s\n", _track.text()));
	} else
	if (isFrame(buf, "TYE")) {
		_year=getID3v2Text(buf+6, size-1);
		FXTRACE((10, "ID3v2 year: %s\n", _year.text()));
	} else
	if (isFrame(buf, "TPB")) {
		_publisher.append(getID3v2Text(buf+6, size-1));
		FXTRACE((10, "ID3v2 publisher: %s\n", _publisher.text()));
	} else
	if (isFrame(buf, "TCO")) {
		FXString g=getID3v2Text(buf+6, size-1);
		_genre=(g.find('(')>=0)?FXString(::genre[FXUIntVal(g.after('(').before(')'))]):g;
		FXTRACE((10, "ID3v2 genre: %s\n", _genre.text()));
	} else
	if (isFrame(buf, "PIC")) {
/*Attached picture   "PIC"
     Frame size         $xx xx xx
     Text encoding      $xx
     Image format       $xx xx xx
     Picture type       $xx
     Description        <textstring> $00 (00)
     Picture data       <binary data>*/
		FXString mime=FXString(buf+11);
		FXTRACE((10, "ID3v2 pic mime: %s\n", mime.text()));
		int i=mime.length();
		char type=buf[12+i];
		FXTRACE((10, "Pic type: %d\n", type));
		const FXString description=FXString(buf+13+i);
		FXTRACE((10, "Pic description: %s\n", description.text()));
		const int j=description.length();
		const int imgOffset=14+i+j;
		FXTRACE((10, "imgOffset: %d\n", imgOffset));
		_img=new imageData(size-imgOffset, (const FXuchar*)buf+imgOffset, mime);
	} else
	/*if (isFrame(buf, "COM")) {
		int i=10;
		while (*(buf+i))
			i++;
		_comment=getID3v2Text(buf+i+1, size-i+9, buf[6]);FIXME
		FXTRACE((10, "ID3v2 comment: %s\n", _comment.text()));
	} else*/
		FXTRACE((10, "ID3v2 unhandled frame %.3s\n", buf));

	return size;
}

int tag::getID3v2Frame(const char* buf, int ver) {

	const FXint size=getFrameSize(buf+4, (ver==4));
	FXTRACE((10, "ID3v2 Frame: %.4s, size: %d\n", buf, size));

	// Flags %0h00kmnp
	// h - Grouping identity
	// k - Compression
	// m - Encryption
	// n - Unsynchronisation
	// p - Data length indicator
	const char flags=buf[9];

	if (flags&0x0E)
		return size;

	// Grouping identity
	if (flags&0x40)
		buf++;

	// Data length indicator
	if (flags&0x01)
		buf+=4;

	if (isFrame(buf, "TIT2")) {
		_title=getID3v2Text(buf+10, size-1);
		FXTRACE((10, "ID3v2 title: %s\n", _title.text()));
	} else
	if (isFrame(buf, "TPE1")) {
		_artist=getID3v2Text(buf+10, size-1);
		FXTRACE((10, "ID3v2 artist: %s\n", _artist.text()));
	} else
	if (isFrame(buf, "TALB")) {
		_album=getID3v2Text(buf+10, size-1);
		FXTRACE((10, "ID3v2 album: %s\n", _album.text()));
	} else
	if (isFrame(buf, "TRCK")) {
		_track=(getID3v2Text(buf+10, size-1));
		FXTRACE((10, "ID3v2 track: %s\n", _track.text()));
	} else
	if (isFrame(buf, "TYER")) {
		_year=getID3v2Text(buf+10, size-1);
		FXTRACE((10, "ID3v2 year: %s\n", _year.text()));
	} else
	if (isFrame(buf, "TPUB")) {
		_publisher=getID3v2Text(buf+10, size-1);
		FXTRACE((10, "ID3v2 publisher: %s\n", _publisher.text()));
	} else
	if (isFrame(buf, "TCON")) {
		FXString g=getID3v2Text(buf+10, size-1);
		_genre=(g.find('(')>=0)?FXString(::genre[FXUIntVal(g.after('(').before(')'))]):g;
		FXTRACE((10, "ID3v2 genre: %s\n", _genre.text()));
	} else
	/*if (isFrame(buf, "COMM")) {
		int i=14;
		while (*(buf+i))
			i++;
		_comment=getID3v2Text(buf+i+1, size-i+9, buf[10]);FIXME
		FXTRACE((10, "ID3v2 comment: %s\n", _comment.text()));
	} else*/
	if (isFrame(buf, "APIC")) {
		FXString mime=FXString(buf+11);
		FXTRACE((10, "ID3v2 pic mime: %s\n", mime.text()));
		const int i=mime.length();
		const char type=buf[12+i];
		FXTRACE((10, "Pic type: %d\n", type));
		const FXString description=FXString(buf+13+i);
		FXTRACE((10, "Pic description: %s\n", description.text()));
		const int j=description.length();

		const int imgOffset=14+i+j;
		FXTRACE((10, "imgOffset: %d\n", imgOffset));

		_img=new imageData(size-imgOffset, (const FXuchar*)buf+imgOffset, mime);
	} else
		FXTRACE((10, "Unhandled frame %.4s\n", buf));

	return size;
}

int tag::getExtHdrSize(const char* buf, int ver) {

	switch (ver) {
	default:
	case 2:
		return 0;
	case 3:
		return getFrameSize(buf, false);
	case 4:
		return getFrameSize(buf, true);
	}
}

bool tag::checkID3v2(char* buf, bool prepend) {

	if (!isFrame(buf, prepend?"ID3":"3DI"))
		return false;

	FXTRACE((10, "ID3v2 tag found at the %s of file\n", prepend?"beginning":"end"));
	return true;
}

bool tag::parseID3v2header(char* buf, int& version, int& tagSize) {

	version=(int)buf[3];
	if (version<2 || version>4) {
		FXTRACE((10, "ID3v2 tag version 2.%d not supported\n", version));
		return 0;
	}

	// Experimental or broken tag
	if (buf[5]&0x3F) {
		FXTRACE((10, "Experimental or broken tag\n"));
		return 0;
	}

	tagSize=getFrameSize(buf+6, true);
	FXTRACE((10, "ID3v2 tag size: %d\n", tagSize));

	const bool unsynch=buf[5]&0x80;
	FXTRACE((10, "ID3v2 unsynch: %d\n", unsynch));
	if (unsynch)
		return 0;

	//bool footer=buf[5]&0x10;

	return buf[5]&0x40;
}

/******** APE ********/

int tag::getLE32(const char* frame) {

	return (FXuint)(FXuchar)frame[0]|(FXuint)(FXuchar)frame[1]<<8|
		(FXuint)(FXuchar)frame[2]<<16|(FXuint)(FXuchar)frame[3]<<24;
}

bool tag::getComment(const char* orig, FXString* dest, const char* type, int& len) {

	int i=strlen(type);
	if (comparecase(orig, type, i))
		return false;

	i++;
	const FXUTF8Codec codec;
	*dest=codec.mb2utf(orig+i, len);
	len+=i;
	FXTRACE((10, "%s: %s\n", type, dest->text()));
	return true;
}

/* Cover art spec
<length> 32 bit
<flags with binary bit set> 32 bit
<field name> "Cover Art (Front)"|"Cover Art (Back)"
0x00
<description> UTF-8 string (needs to be a file name to be recognized by AudioShell - meh)
0x00
<cover data> binary
*/
int tag::getAPEItem(const char* buf) {

	FXint itemSize=getLE32(buf);
	//FXint flags=getLE32(buf+4);

	const char *ptr=buf+8;
	if (!getComment(ptr, &_title, "title", itemSize))
	if (!getComment(ptr, &_artist, "artist", itemSize))
	if (!getComment(ptr, &_year, "year", itemSize))
	if (!getComment(ptr, &_album, "album", itemSize))
	if (!getComment(ptr, &_genre, "genre", itemSize))
	if (!getComment(ptr, &_comment, "comment", itemSize)) {
		if (!comparecase(ptr, "track", 5)) {
			_track=FXString(ptr+6);
			itemSize+=6;
		}
	}

	return itemSize+8;
}

bool tag::checkAPE(char* buf, int& itemsSize, int& tagSize) {

	if (!isFrame(buf, "APETAGEX"))
		return false;

	FXTRACE((10, "APE tag found\n"));

	const int version=getLE32(buf+8);
	FXTRACE((10, "APE tag version: %d\n", version));

	tagSize=getLE32(buf+12);
	FXTRACE((10, "APE tag size: %d\n", tagSize));

	FXTRACE((10, "APE tag items: %d\n", getLE32(buf+16)));

	itemsSize=tagSize-APE_HEADER_SIZE;

	if (version>1000) {
		FXint flags=getLE32(buf+20);

		if (flags&0x80000000)
			tagSize+=APE_HEADER_SIZE;
	}

	return true;
}
