/*
 *  Copyright (C) 2011 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "xdg.h"

#ifdef _WIN32
#  include <windows.h>
#  include <shlobj.h>
#endif

#ifdef _WIN32
const FXString xdg::getWindowsDir(const int csidl) {

	TCHAR szPath[MAX_PATH];

	if (SHGetSpecialFolderPath(NULL, szPath, csidl, 1))
		return FXString(szPath);
	else
		return FXString::null;
}
#endif

const FXString xdg::getCacheDir(void) {
#ifdef _WIN32
	return getWindowsDir(CSIDL_INTERNET_CACHE);
#else
	FXString xdgCacheDir=FXSystem::getEnvironment("XDG_CACHE_HOME");
	if (xdgCacheDir.empty())
		xdgCacheDir=FXSystem::getHomeDirectory()+"/.cache";
	return FXPath::expand(xdgCacheDir);
#endif
}

const FXString xdg::getRuntimeDir(void) {
#ifndef _WIN32
	FXString xdgRuntimeDir=FXSystem::getEnvironment("XDG_RUNTIME_DIR");
	if (!xdgRuntimeDir.empty())
		return FXPath::expand(xdgRuntimeDir);
#endif
	return getCacheDir();
}

const FXString xdg::getConfigDir(void) {
#ifdef _WIN32
	return getWindowsDir(CSIDL_APPDATA);
#else
	FXString xdgConfigDir=FXSystem::getEnvironment("XDG_CONFIG_HOME");
	if (xdgConfigDir.empty())
		xdgConfigDir=FXSystem::getHomeDirectory()+"/.config";
	return FXPath::expand(xdgConfigDir);
#endif
}

#define MAXBUFFER 2048

const char* xdg::parseFile(FXFile& file, char *buffer, FXint& len, FXint& end) {

	if (end==0) {
		end=file.readBlock(buffer, MAXBUFFER);
		if (end<=0)
			return 0;
	}

	FXint pos=len;
	FXint i=pos;
	while (i<end && buffer[i]!='\n') ++i;
	if (i<end) {
		buffer[i]=0;
		len=i+1;
	} else {
		end=0;
		len=0;
		return 0;
	}

	return buffer+pos;
}

const FXString xdg::getMusicDir(void) {
#ifdef _WIN32
	return getWindowsDir(CSIDL_MYMUSIC);
#else
	FXString xdgUserDirs=getConfigDir();
	xdgUserDirs.append("/user-dirs.dirs");
	FXTRACE((10, "xdgUserDirs: %s\n", xdgUserDirs.text()));

	FXString musicDir;
	FXFile userDirs(xdgUserDirs, FXIO::Reading);
	if (userDirs.isOpen()) {
		char buffer[MAXBUFFER];
		FXint eol=0,end=0;
		const char* line;
		while (line=parseFile(userDirs, buffer, eol, end)) {
			FXTRACE((10, "user-dirs.dirs: %s\n", line));
			if (!compare(line, "XDG_MUSIC_DIR", 13))
				musicDir=FXString(line).after('=').substitute('"', ' ').trim();
		}
	}

	return FXPath::expand(musicDir);
#endif
}

bool xdg::open(const FXString& link) {
#ifdef _WIN32
#ifdef UNICODE
	FXint n=ncslen(link.text(), link.length());
	FXnchar *buffer=new FXnchar[n+1];
	utf2ncs(buffer,link.text(), link.length()+1);
	n=((int)ShellExecute(NULL, L"open", buffer, NULL, NULL, SW_SHOW)>32);
	delete [] buffer;
	return n;
#else
	return ((int)ShellExecute(NULL, "open", link.text(), NULL, NULL, SW_SHOW)>32);
#endif // UNICODE
#else
	const FXString xdgopen=FXPath::search(FXSystem::getExecPath(), "xdg-open");
	if (!xdgopen.empty() && FXStat::isExecutable(xdgopen))
		return (system(FXStringFormat("xdg-open \"%s\" >/dev/null 2>&1 & ", link.text()).text())>=0);

	return false;
#endif
}
